// <file>
//     <license see="prj:///doc/license.txt"/>
//     <owner name="Stefan Wallner" />
// </file>

using System;
using System.Text;
using System.IO;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.Reflection;
using System.Windows.Forms;
using System.CodeDom.Compiler;
using System.Collections.Generic;

using ICSharpCode.Core;
using ICSharpCode.SharpDevelop;
using ICSharpCode.SharpDevelop.Project;
using ICSharpCode.SharpDevelop.Gui;
using ICSharpCode.TextEditor;
using ICSharpCode.TextEditor.Document;		//IDocument
using ICSharpCode.SharpDevelop.DefaultEditor.Gui.Editor;

using CocoAddin.Tools;

namespace CocoAddin.Commands
{
	/// <summary>
	/// command to start Coco/R compiling process
	/// </summary>
	public class RunCoco : AbstractCommand
	{
		
		//manages the Coco/R compiler with all his settings
		private CocoCompilerManager coco;
		
		//active editor window
		private IWorkbenchWindow activeWindow = null;
		
		private string outputType = null;
		private MSBuildProject MSproj = null; 
		
		/// <summary>
		/// start Coco/R compiler if possible
		/// </summary>
		public override void Run()
		{ 
			try {
				CocoPad pad = (CocoPad)Owner;
				CSharpCompilerResult result;	
				IProject proj = ProjectService.CurrentProject;
				
				coco = new CocoCompilerManager();
				//the atg-file we want to compile
				coco.properties.InputFile = CocoPad.ActiveParsedFile;
				
				//clear all previous output
				TaskService.BuildMessageViewCategory.ClearText();
				TaskService.Clear();				
	
				if (coco.properties.InputFile == null || coco.properties.InputFile.Length == 0){
 					MessageService.ShowMessage("Please choose an ATG-file !");
 					return;
 				}
 				
 				if (proj == null){
 					MessageService.ShowMessage("Please select a project with at least one ATG-File !");
 					return;
 				}
				
				//if window is not open
				if (activeWindow == null)
					FileService.OpenFile(coco.properties.InputFile);
				
				activeWindow = WorkbenchSingleton.Workbench.ActiveWorkbenchWindow;
				
				//try to find the frame files directory
				string atgDir = Path.GetDirectoryName(CocoPad.ActiveParsedFile);
				if (coco.properties.CocoFrame == ""){
	 				Console.WriteLine("Nichts definiert !");
		 			if (proj.IsFileInProject(atgDir + Path.DirectorySeparatorChar+"Parser.frame") && 
		 			    proj.IsFileInProject(atgDir + Path.DirectorySeparatorChar+"Scanner.frame")) {
		 				coco.properties.CocoFrame = atgDir;
		 			} else if (FileUtility.TestFileExists(atgDir + Path.DirectorySeparatorChar+"Parser.frame") && 
		 			           FileUtility.TestFileExists(atgDir + Path.DirectorySeparatorChar+"Scanner.frame")) {
		 				coco.properties.CocoFrame = atgDir;
		 			} else {
		 				MessageService.ShowMessage("No frame files found in the actual project and in the directory of your actual ATG-File !\n" +
		 				                           "No frame files directory defined in Coco/R Options !\n" +
		 				                           "Using the frame files in: "+coco.properties.PluginDir);
		 				coco.properties.CocoFrame = coco.properties.PluginDir;
		 			}
 				} 
			
				string parserFile = Path.Combine(coco.properties.OutPutPath, "Parser.cs"); 
				string scannerFile = Path.Combine(coco.properties.OutPutPath, "Scanner.cs");
				string traceFile = Path.Combine(coco.properties.OutPutPath, "trace.txt");
				
				//close eventually open Parser.cs, Scanner.cs or trace.txt files
				List<IWorkbenchWindow> wtoDelete = new List<IWorkbenchWindow>(3);
				foreach(IViewContent c in WorkbenchSingleton.Workbench.ViewContentCollection){
 					if (FileUtility.IsEqualFileName(c.FileName, parserFile) || FileUtility.IsEqualFileName(c.FileName, scannerFile) || 
					    FileUtility.IsEqualFileName(c.FileName, traceFile)){
							wtoDelete.Add(c.WorkbenchWindow);
	 				}
	 			}
				foreach (IWorkbenchWindow w in wtoDelete){
					w.CloseWindow(true);
				}
							
				//clear eventually previous generated files in the project
				List<ProjectItem> toDelete = new List<ProjectItem>();
					for (int i=0; i<proj.Items.Count; ++i){
	 					ProjectItem projItem = proj.Items[i] as ProjectItem;
	 					if (FileUtility.IsEqualFileName(projItem.FileName, parserFile))
	 						toDelete.Add(projItem);
	 						
	 					if (FileUtility.IsEqualFileName(projItem.FileName, scannerFile))
	 						toDelete.Add(projItem);
	 						
	 					if (FileUtility.IsEqualFileName(projItem.FileName, traceFile))
	 						toDelete.Add(projItem);
	 				}
	 				foreach(ProjectItem i in toDelete){
	 					LoggingService.Debug("Remove Item: " + i.ToString());
	 					ProjectService.RemoveProjectItem(proj, i);	
	 				}
				proj.Save();
				
				//start Coco/R
 				result = coco.Compile();
 				if (result == null) {
 					MessageService.ShowWarning("No valid Coco/R Compiler Generator found !\nTry to choose a valid Coco/R in Tools -> Options -> Tools -> Coco/R !");
 					return;
 				}
 				if (result.CompilerResults.Errors.Count > 0) {
 					//bring errors from Coco/R to output
	 				foreach (CompilerError cError in result.CompilerResults.Errors){ 
	 						TaskService.Add(new Task(cError.FileName, cError.ErrorText, cError.Column-1, cError.Line-1, TaskType.Error, proj));
 					}
 					TaskService.BuildMessageViewCategory.AppendText(result.CompilerOutput);
 					TaskService.BuildMessageViewCategory.AppendText("\nCoco/R Addin ("+Assembly.GetExecutingAssembly().GetName().Version+")");
 					TaskService.BuildMessageViewCategory.AppendText("\nSemantic action errors not checked because there were Coco/R errors !");
	 				WorkbenchSingleton.Workbench.GetPad(typeof(CompilerMessageView)).RedrawContent();
	 				WorkbenchSingleton.Workbench.GetPad(typeof(ErrorListPad)).BringPadToFront();
	 				return;
 				} else {
 					//no errors from Coco/R find semantic action errors
 					TaskService.BuildMessageViewCategory.AppendText(result.CompilerOutput);
	
 					//put the generated parser and scanner file to the project
 					PadDescriptor pd = WorkbenchSingleton.Workbench.GetPad(typeof(ProjectBrowserPad));
 			
 					ProjectItem item = ProjectItemFactory.CreateProjectItem(proj,"Compile");
 					item.FileName = coco.properties.OutPutPath + Path.DirectorySeparatorChar + "Parser.cs";
 					ProjectService.AddProjectItem(proj, item);
 					item = ProjectItemFactory.CreateProjectItem(proj,"Compile");
 					item.FileName = coco.properties.OutPutPath + Path.DirectorySeparatorChar + "Scanner.cs";
 					ProjectService.AddProjectItem(proj, item);
	 				if (!coco.properties.TraceString.Equals("")){
 						item = ProjectItemFactory.CreateProjectItem(proj,"Content");
 						item.FileName = coco.properties.OutPutPath + Path.DirectorySeparatorChar + "trace.txt";
 						ProjectService.AddProjectItem(proj, item);
	 				}
 					
	 				TaskService.BuildMessageViewCategory.AppendText("\nCoco/R Addin ("+Assembly.GetExecutingAssembly().GetName().Version+")");
	 				
	 				//try to compile in background and search for errors in the semantic actions
	 				MSproj = (MSBuildProject) proj;
	 				//save the actual OutputType
	 				outputType = MSproj.GetProperty("OutputType");
	 				MSproj.SetProperty("OutputType","Library");
	 				proj.Save();
	 				
	 				if (ProjectBrowserPad.Instance != null){
 						ProjectBrowserPad.Instance.ProjectBrowserControl.RefreshView();
 						ProjectBrowserPad.Instance.ProjectBrowserControl.TreeView.ExpandAll();
 					}
	 				
	 				//start compiling with C# compiler
	 				proj.Build(BuildFinished);
 				} 	
			} catch (CocoAddinException ex) {
				MessageService.ShowMessage(ex.Message);
			}
		}//end Run
		
		
		/// <summary>
		/// handles the compiler Errors of the C# Compiler an filter semantic errors
		/// </summary>
		void BuildFinished(CompilerResults csharpresult){
			
			//possible errors in the semantic actions
			if (csharpresult.Errors.HasErrors) {	
				TextEditorControl textArea = ((ITextEditorControlProvider)activeWindow.ViewContent).TextEditorControl;
				Hashtable hash = this.generateSemanticHashtable(textArea.Document);
				
				StreamReader parser = new StreamReader(coco.properties.OutPutPath + Path.DirectorySeparatorChar + "Parser.cs");
				StreamReader scanner = new StreamReader(coco.properties.OutPutPath + Path.DirectorySeparatorChar + "Scanner.cs");
				string tmp = "";
				int pLine = 0,sLine = 0,sError = 0;
				
				foreach (CompilerError cError in csharpresult.Errors){
					CompilerError atgError;
					
					if (cError.FileName.EndsWith("Parser.cs")) {
						while (pLine < cError.Line) {
							tmp = parser.ReadLine();
							pLine++;
						}
					} else if (cError.FileName.EndsWith("Scanner.cs")) {
						while (sLine < cError.Line) {
							tmp = scanner.ReadLine();
							sLine++;
						}
					} else {//no scanner or parser error -> skip this error
						continue;
					}
					
					//do we have an semantic action error ?
					if (hash.ContainsKey(tmp.Trim())) {
						atgError = (CompilerError)hash[tmp.Trim()];
						cError.FileName = coco.properties.InputFile;
						cError.Line = atgError.Line+1;
						cError.Column = atgError.Column + cError.Column - this.getLineStart(tmp,0) - 1;
						TaskService.Add(new Task(cError));
						TaskService.BuildMessageViewCategory.AppendText("\n-- line "+cError.Line+" col "+cError.Column+": "+cError.ErrorText);
						sError++;
					}
				}
				
				
				TaskService.BuildMessageViewCategory.AppendText("\n"+sError+" semantic action errors detected.");
				//errors outside the semantic actions
				if (csharpresult.Errors.Count >= sError) {
					TaskService.BuildMessageViewCategory.AppendText("\n" + (csharpresult.Errors.Count-sError) + " errors or warnings in generated files detected.");
					WorkbenchSingleton.Workbench.GetPad(typeof(CompilerMessageView)).RedrawContent();
					WorkbenchSingleton.Workbench.GetPad(typeof(CompilerMessageView)).BringPadToFront();
				} else {
					WorkbenchSingleton.Workbench.GetPad(typeof(CompilerMessageView)).RedrawContent();
					WorkbenchSingleton.Workbench.GetPad(typeof(ErrorListPad)).BringPadToFront();
				}
				
				parser.Close();
				scanner.Close();
			} else {	//no Errors detected
				FileService.OpenFile(coco.properties.OutPutPath + Path.DirectorySeparatorChar + "Scanner.cs");
				FileService.OpenFile(coco.properties.OutPutPath + Path.DirectorySeparatorChar + "Parser.cs");
				TaskService.BuildMessageViewCategory.AppendText("\n0 semantic action errors detected");
				WorkbenchSingleton.Workbench.GetPad(typeof(CompilerMessageView)).BringPadToFront();
			}
			
			MSproj.SetProperty("OutputType", outputType);
 			ProjectService.CurrentProject.Save();
			
		}
	
 		/// <summary>
 		/// generates a hashtable
 		/// keys are the trimmed codelines in the semantic actions
 		/// in the hashtable are CompilerError elements with the line of the semantic action and
 		/// the starting column of the codeline int the semantic action
 		/// </summary>
 		private Hashtable generateSemanticHashtable(IDocument doc){
 			Hashtable h = new Hashtable();
 			bool semanticMode=false;
 			
 			LoggingService.Debug("start generateHashtable");
 			if (doc == null) return null;
 			
 			for (int line = 0; line < doc.TotalNumberOfLines; line++){
 				LineSegment lineSeg = doc.GetLineSegment(line);
 				
 				List<TextWord> wList = lineSeg.Words;
 				
 				string read = "";
 				int beg = -1;
 				int end = -1;
 				int startSem = 0;
 				
 				if (wList == null) continue;
 					
 				//switches to semantic mode if we have a semantic action in the current line
 				for (int i=0; i<wList.Count-1; i++){	
 					TextWord first = (TextWord)wList[i];
 					TextWord next = (TextWord)wList[i+1]; 					 					
 					
 					if (!first.IsWhiteSpace && first.Length==1 && first.Word[0]=='(' && 					
 						!next.IsWhiteSpace && next.Length==1 && next.Word[0]=='.'){
 							beg = next.Offset;
 							semanticMode = true;
 							continue;
 					} 					 
 					if (!first.IsWhiteSpace && first.Length==1 && first.Word[0]=='.' && 
 						!next.IsWhiteSpace && next.Length == 1 && next.Word[0]==')'){
 							end = next.Offset;
 					}
 				}
 				
 				
 				if (semanticMode){
 					
	 				if ((beg > -1) && (end > -1)) { 
	 					read = doc.GetText(lineSeg.Offset + beg +1, end - beg -2);
	 					startSem = TextUtilities.GetFirstNonWSChar(doc, lineSeg.Offset + beg + 1)-lineSeg.Offset+1;
	 					semanticMode = false;
	 					//LoggingService.Debug("ONE :\t"+ startSem+" ("+line+","+beg+","+end+")"+"read:\t \"" +read+"\"");	
	 				} else if(beg > -1) {
	 					read = doc.GetText(lineSeg.Offset + beg +1, lineSeg.Length - beg - 1);
	 					startSem = TextUtilities.GetFirstNonWSChar(doc, lineSeg.Offset + beg + 1)-lineSeg.Offset+1;
	 					//LoggingService.Debug("BEGIN :\t"+ startSem+" ("+line+","+beg+","+end+")"+"read:\t \"" +read+"\"");	
	 				} else if(end > -1) {
 						semanticMode = false;
 						read = doc.GetText(lineSeg.Offset, end - 2);
 						startSem = TextUtilities.GetFirstNonWSChar(doc, lineSeg.Offset)-lineSeg.Offset+1;
 						if (startSem == end) continue;
	 					//LoggingService.Debug("END :\t"+ startSem+" ("+line+","+beg+","+end+")"+"read:\t \"" +read+"\"");
 					} else {
 						read = doc.GetText(lineSeg.Offset, lineSeg.Length);	
 						startSem = (TextUtilities.GetFirstNonWSChar(doc, lineSeg.Offset))-lineSeg.Offset+1;
 						//LoggingService.Debug("MID :\t"+ startSem+" ("+line+","+beg+","+end+")"+"read:\t \"" +read+"\""); 		
 					}
	 				
	 				
	 				try {
	 						CompilerError cError = new CompilerError();
		 					cError.Line = line;
		 					cError.Column = startSem; 
		 					h.Add(read.Trim(), cError);
		 					
	 					} catch(ArgumentException) {
 							LoggingService.Debug("line duplicated ! : "+read.Trim()+"\"");
	 					}
 				}
 			}
 			return h;
 		}
 		
 		/// <summary>
 		/// get the first non whitespace carachter location in the stringline
 		/// </summary>
 		private int getLineStart(string tmp, int start) {
 			int back = 0;
 			char[] ch = tmp.ToCharArray();
 			
 			for(int i=start; i<ch.Length; i++) {
 				if (!Char.IsWhiteSpace(ch[i]))
 					break;
 				back++;
 			}
 			return back;
 		}
	}
}
